;; Generated package description from futur.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "futur" "1.7.0.20260721.9" "Future/promise-based async library" 'nil :commit "e307e22054d2d3f64ad9cfe998dd2b33cfc16ca7" :url "https://elpa.gnu.org/packages/futur.html" :authors '(("Stefan Monnier" . "monnier@iro.umontreal.ca")) :maintainer '("Stefan Monnier" . "monnier@iro.umontreal.ca") :keywords '("concurrency" "async" "promises" "futures"))
